from .utils.advanced_pickling import map_pickling
from .utils.locked import locked
from .utils.optional_context import optional_context
from .shared_dict import SharedDictDescriptor
from .types import ThreadPool, ProcessPool, ThreadManager, ProcessManager

import functools

from typing import Union, Optional

__all__ = ['multiprocess_foreach', 'multithread_foreach']


def _foreach(do, iterator, shared_dict_descriptor: Optional[SharedDictDescriptor],
             pool_builder: Union[ThreadPool, ProcessPool],
             manager_builder: Union[ThreadManager, ProcessManager]):
    # FIXME(ekon): if chuncksize is bigger than len(iterator),
    #              the function does nothing

    # TODO(ekon): build in tqdm

    # TODO(ekon): set the shared_dict struct using an arg

    # TODO(ekon): make it possible to choose between map, imap, imap_unordered etc.

    # If swap these context managers, you will probably
    # face with a file not found error
    # Read this:
    # https://stackoverflow.com/questions/56641428/python-3-6-nested-multiprocessing-managers-cause-filenotfounderror

    optional_manager = lambda: optional_context(bool(shared_dict_descriptor), manager_builder)
    with optional_manager() as manager:
        if manager:
            # TODO(ekon): pass a descriptor and 'manager' to SharedDict
            shared_dict = manager.dict()
            # with shared_dict['locked']:
            #     ...
            shared_dict['locked'] = functools.partial(locked, manager.Lock())
        else:
            shared_dict = None

        with pool_builder() as pool:
            special_map_pickling: tuple = map_pickling(
                worker_func=do,
                items=iterator,
                shared_dict=shared_dict,
            )

            results = pool.imap_unordered(
                *special_map_pickling,
                chunksize=chuncksize,
            )

            # If we do not get rid of the generator here, within the pool context,
            # we won't be able to do it anywhere else
            results = results if type(results) is list else list(results)
    return results


multiprocess_foreach = functools.partial(_foreach, pool_builder=ProcessPool, manager_builder=ProcessManager)
multithread_foreach = functools.partial(_foreach, pool_builder=ThreadPool, manager_builder=ThreadManager)
