from foreach.utils.locked import locked
from .types import AnyManager

import functools

from typing import Union
from dataclasses import dataclass


@dataclass
class SharedDictDescriptor:
    items: Union[list, dict]
    need_locker: bool

    def build(self, manager: AnyManager):
        sd = manager.dict()
        if self.need_locker:
            sd['locker'] = functools.partial(locked, manager.Lock())
        if type(self.items) is list:
            sd['data'] = {}.fromkeys(self.items, None)
        elif type(self.items) is dict:
            sd['data'] = self.items
        return sd


# TODO(ekon):
"""
Redefine print in worker to use lock. All user has to do is to use `print`
def new_print():
    with self.locker:
        print(*args, **kwargs)
        
print = new_print
"""


class SharedDictWrapper:
    def __init__(self, manager: AnyManager, descriptor: SharedDictDescriptor):
        self._shared_dict = descriptor.build(manager)

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        if 'locker' not in self._shared_dict:
            raise KeyError('Your SharedDict does not use locker')
        with self.locker:
            self._data[key] = value

    @property
    def locker(self):
        return self._shared_dict['locker']

    @property
    def _data(self):
        return self._shared_dict['data']
