from contextlib import contextmanager
from typing import Protocol
from abc import abstractmethod


class LockLike(Protocol):
    @abstractmethod
    def acquire(self):
        pass

    @abstractmethod
    def release(self):
        pass


@contextmanager
def locked(lock: LockLike):
    """
        >>> def worker(nums, shared_dict: dict,nums=None):
        ...     lock = shared_dict['lock']
        ...     with locked(lock):
        ...         shared_dict['called'] += 1
        ...     return [n**2 for n in nums]
    """
    lock.acquire()
    yield
    lock.release()
